#include <linux/device.h>
#include <linux/gpio.h>
#include <linux/vgpio.h>
#include <linux/delay.h>
#include <linux/clk.h>

#include <linux/amba/bus.h>
#include <linux/amba/bcmring_clcd.h>

#include <plat/irvine.h>
#include <plat/irvine-lcd.h>

#include <mach/pinmux.h>

#define LCM_MODEL_NAME  "LMS606KF01"

#define HBP     (8)
#define HFP     (8)
#define HSW     (10)
#define PPL     (480)

#define VBP     (8)
#define VFP     (8)
#define VSW     (2)
#define LPP     (800)

#define PCD_HI  (0)
#define BCD     (0)
#define CPL     (PPL-1)
#define IOE     (0)
#define IPC     (1)
#define IHS     (1)
#define IVS     (1)
#define ACB     (0)
#define CLKSEL  (0)
#define PCD_LO  (3)

//#define LMS606KF01_DEBUG

#ifdef LMS606KF01_DEBUG

static void lms606kf01_check_commands(void);
static void lms606kf01_dump_level1(void);

enum com_type {
	R,
	W,
	C	
};

struct lms606kf01_level1_command {
	unsigned char opcode;
	const char*   str;
	enum com_type rwc;
	int			  para_size;
};

static struct lms606kf01_level1_command level1_commands []= {
	{0x00,	"nop"	,C	,0	},
	{0x01,	"soft_reset"	,C	,0	},
	{0x04,	"read_display_identification_information"	,R	,3	},
	{0x06,	"get_red_channel"	,R	,1	},
	{0x07,	"get_green_channel"	,R	,1	},
	{0x08,	"get_blue_channel"	,R	,1	},
	{0x0A,	"get_power_mode"	,R	,1	},
	{0x0B,	"get_address_mode"	,R	,1	},
	{0x0C,	"get_pixel_format"	,R	,1	},
	{0x0D,	"get_display_mode"	,R	,1	},
	{0x0E,	"get_signal_mode"	,R	,1	},
	{0x0F,	"get_diagnostic_result"	,R	,1	},
	{0x10,	"enter_sleep_mode"	,C	,0	},
	{0x11,	"exit_sleep_mode"	,C	,0	},
	{0x12,	"enter_partial_mode"	,C	,0	},
	{0x13,	"enter_normal_mode"	,C	,0	},
	{0x20,	"exit_invert_mode"	,C	,0	},
	{0x21,	"enter_invert_mode"	,C	,0	},
	{0x26,	"set_gamma_curve"	,W	,1	},
	{0x28,	"set_display_off"	,C	,0	},
	{0x29,	"set_display_on"	,C	,0	},
	{0x2A,	"set_column_address"	,W	,4	},
	{0x2B,	"set_page_address"	,W	,4	},
	{0x2C,	"write_memory_start"	,W	,-1	},
	{0x2D,	"write_LUT"	,W	,-1	},
	{0x2E,	"read_memory_start"	,R	,-1	},
	{0x30,	"set_partial_area"	,W	,4	},
	{0x33,	"set_scroll_area"	,W	,6	},
	{0x34,	"set_tear_off"	,C	,0	},
	{0x35,	"Set_tear_on"	,W	,1	},
	{0x36,	"set_address_mode"	,W	,1	},
	{0x37,	"set_scroll_start"	,W	,2	},
	{0x38,	"exit_idle_mode"	,C	,0	},
	{0x39,	"enter_idle_mode"	,C	,0	},
	{0x3A,	"set_pixel_format"	,W	,1	},
	{0x3C,	"write_memory_continue"	,W	,-1	},
	{0x3E,	"read_memory_continue"	,R	,-1	},
	{0x44,	"set_tear_scanline"	,W	,2	},
	{0x45,	"get_scanline"	,R	,2	},
	{0xA1,	"read_DDB_start"	,R	,5	},
	{0xA8,	"read_DDB_continue"	,R	,-1	},
	{0xDA,	"read_ID1"	,R	,1	},
	{0xDB,	"read_ID2"	,R	,1	},
	{0xDC,	"read_ID3"	,R	,1	},
};

#define LMS606KF01_CHECK(x) \
do{\
	unsigned char buf[64]; \
	int i , len; \
	len = sizeof(x)/sizeof(x[0]) - 1; \
	level2_spi_read(x[0], buf , len); \
	for ( i = 0; i < len; i++) \
		if (x[1+i] != buf[i]) \
			printk(KERN_ERR LCM_MODEL_NAME " cmd(0x%x): %d Para, Set:0x%x, Get:0x%x \n", x[0], 1+i, x[1+i], buf[i]);\
}while(0)

#define LMS606KF01_FUNC_TRACE printk(KERN_INFO LCM_MODEL_NAME " %s() \n", __FUNCTION__ );
#else 
#define LMS606KF01_FUNC_TRACE
#endif //#ifdef LMS606KF01_DEBUG


#define LMS606KF01_SPI_DELAY 40

#define LMS606KF01_SET(x) \
do{\
	spi_write_set(x, (sizeof(x) / sizeof(x[0])));\
}while(0)


static unsigned char password[] = 
{ 
	0xF0, 0x5A, 0x5A
};

static unsigned char vcom[] = 
{ 
	0xF4, 0x30, 0x3F, 0x5E
};

static unsigned char smps_posi[] = 
{ 
	0xBD, 0x06, 0x0B, 0x7E, 0x03,
	0x12, 0x37
};

static unsigned char smps_nega[] = 
{ 
	0xBE, 0x06, 0x0B, 0x7E, 0x03,
	0x15, 0x37
};

static unsigned char smps[] = 
{ 
	0xBF, 0x02, 0x0F
};

static unsigned char power_control[] = 
{ 
	0xF3, 0x10, 0xA9, 0x00, 0x01, 
	0x44, 0xA4, 0x83, 0x83, 0x00,
	0x50
};

static unsigned char source_control[] = 
{ 
	0xF5, 0x03, 0x09, 0x80, 0x01,
	0x01, 0x00, 0x00, 0x05, 0x03
};

static unsigned char positive_r_gamma[] = 
{ 
	0xE7, 0x05, 0x35, 0x1E, 0x1C,
	0x1B, 0x1D, 0x18, 0x1F, 0x1F,
	0x24, 0x2C, 0x36, 0x36, 0x37,
	0x28, 0x16, 0x00
};

static unsigned char positive_g_gamma[] = 
{ 
	0xE8, 0x05, 0x35, 0x1E, 0x1C,
	0x1B, 0x1D, 0x18, 0x22, 0x1F,
	0x27, 0x2E, 0x39, 0x3C, 0x2D,
	0x23, 0x08, 0x0C
};

static unsigned char positive_b_gamma[] = 
{ 
	0xE9, 0x05, 0x35, 0x1E, 0x1C,
	0x1B, 0x1D, 0x18, 0x1F, 0x21,
	0x28, 0x2E, 0x3A, 0x34, 0x2D,
	0x28, 0x0A, 0x00
};

static unsigned char negative_r_gamma[] = 
{ 
	0xEA, 0x05, 0x23, 0x0A, 0x0A,
	0x07, 0x02, 0x04, 0x09, 0x0A,
	0x0E, 0x15, 0x10, 0x0E, 0x12,
	0x12, 0x04, 0x00
};

static unsigned char negative_g_gamma[] = 
{ 
	0xEB, 0x05, 0x10, 0x0A, 0x0A,
	0x07, 0x02, 0x04, 0x06, 0x0A,
	0x11, 0x19, 0x15, 0x16, 0x1C,
	0x1D, 0x08, 0x0C
};

static unsigned char negative_b_gamma[] = 
{ 
	0xEC, 0x05, 0x23, 0x0A, 0x0A,
	0x07, 0x02, 0x04, 0x09, 0x0C,
	0x12, 0x1B, 0x18, 0x26, 0x23,
	0x26, 0x14, 0x00
};

static unsigned char display_control[] = 
{ 
	0xF2, 0x01, 0x80, 0x00, 0x04,
	0x08, 0x08
};

static unsigned char resolution[] = 
{ 
	0xB8, 0x12
};

static unsigned char pixel_format[] = 
{ 
	0x3A, 0x50
};

static unsigned char sleep_out[] = 
{ 
	0x11
};

static unsigned char disp_on[] = 
{ 
	0x29
};

static unsigned char disp_off[] = 
{ 
	0x28
};

static unsigned char sleep_in[] = 
{ 
	0x10
};

static unsigned char spi_read_en[] =
{
	0xB1, 0x01
};

static unsigned char spi_read_dis[] =
{
	0xB1, 0x00
};

static int lms606kf01_poweron_from_suspend = 0;

#define SPI_COMMAND	(0x00)
#define SPI_DATA	(0x01)

static inline void spi_write(unsigned char is_data, unsigned char data)
{
	int i;
	
	gpio_set_value(TT_VGPIO_LCD_CS, 0);
	udelay(LMS606KF01_SPI_DELAY);
    if(is_data)
    {
        /* control bit is 1 */
        gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
        udelay(LMS606KF01_SPI_DELAY);
        gpio_set_value (TT_VGPIO_LCD_SDI, 1);
        udelay(LMS606KF01_SPI_DELAY);
        gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
        udelay(LMS606KF01_SPI_DELAY);
    }
    else
    {    
        /* control bit is 0 */
        gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
        udelay(LMS606KF01_SPI_DELAY);
        gpio_set_value (TT_VGPIO_LCD_SDI, 0);
        udelay(LMS606KF01_SPI_DELAY);
        gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
        udelay(LMS606KF01_SPI_DELAY);
    }
            
	for (i = 0; i < 8; i++) {
		gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
		udelay(LMS606KF01_SPI_DELAY);
		
		if (data & 0x80)
			gpio_set_value (TT_VGPIO_LCD_SDI, 1);
		else
			gpio_set_value (TT_VGPIO_LCD_SDI, 0);

		udelay(LMS606KF01_SPI_DELAY);
		gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
		udelay(LMS606KF01_SPI_DELAY);
		data <<= 1;
	}

	//gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
	gpio_set_value(TT_VGPIO_LCD_CS, 1);
    udelay(10);
    gpio_set_value(TT_VGPIO_LCD_SDI, 1);
    udelay(LMS606KF01_SPI_DELAY);
}

#define spi_write_command(x) spi_write(0,x)
#define spi_write_data(x)  spi_write(1,x)

static void spi_write_set(unsigned char * data, int len)
{
	int i;	
    spi_write_command(data[0]);

	for (i = 1; i < len; i++) {
		spi_write_data(data[i]);
	}
}

static void lms606kf01_power_on(void)
{
	LMS606KF01_FUNC_TRACE
	
	mdelay(10);
	gpio_set_value(TT_VGPIO_LCD_ON, 1);
	mdelay(120);

	LMS606KF01_SET(password);
	LMS606KF01_SET(vcom);
	LMS606KF01_SET(sleep_out);
	
	mdelay(1);
	
	LMS606KF01_SET(smps_posi);
	LMS606KF01_SET(smps_nega);
	LMS606KF01_SET(smps);
	LMS606KF01_SET(power_control);
	LMS606KF01_SET(source_control);
	LMS606KF01_SET(positive_r_gamma);
	LMS606KF01_SET(positive_g_gamma);
	LMS606KF01_SET(positive_b_gamma);
	LMS606KF01_SET(negative_r_gamma);
	LMS606KF01_SET(negative_g_gamma);
	LMS606KF01_SET(negative_b_gamma);
	LMS606KF01_SET(display_control);
	LMS606KF01_SET(resolution);
	LMS606KF01_SET(pixel_format);

	mdelay(120);
	
	LMS606KF01_SET(disp_on);	

#ifdef LMS606KF01_DEBUG
	lms606kf01_check_commands();
	lms606kf01_dump_level1();
#endif
}

static void lms606kf01_display_on(void)
{
	LMS606KF01_FUNC_TRACE
	
	/* From cold boot, LCD controller already starts to function right here;
	   from resume, LCD controller does not function yet. */	   
	if(!lms606kf01_poweron_from_suspend)
		lms606kf01_power_on();	
}

static void lms606kf01_display_off(void)
{
	LMS606KF01_FUNC_TRACE
	
	LMS606KF01_SET(disp_off);
	LMS606KF01_SET(sleep_in);

	mdelay(120);

	gpio_set_value (TT_VGPIO_LCD_ON, 0);
	gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
	gpio_set_value (TT_VGPIO_LCD_SDI, 1);
	gpio_set_value (TT_VGPIO_LCD_CS, 1); 
	lms606kf01_poweron_from_suspend = 1;
}

static void lms606kf01_enable (struct clk *clk)
{
	LMS606KF01_FUNC_TRACE

	/*	In cold boot, lms606kf01_enable is called two times.
		we use this flag to skip calling in cold boot. */		
	if(lms606kf01_poweron_from_suspend)
		lms606kf01_power_on();
}

static void lms606kf01_disable (struct clk *clk)
{
	LMS606KF01_FUNC_TRACE
}

static void lms606kf01_suspend (struct clk *clk)
{
	LMS606KF01_FUNC_TRACE
	lms606kf01_display_off();
}

static void lms606kf01_resume (struct clk *clk)
{
	LMS606KF01_FUNC_TRACE
	lms606kf01_display_on();
}

static void lms606kf01_init (struct clk *clk)
{
	LMS606KF01_FUNC_TRACE
	lms606kf01_display_on();
}

static void lms606kf01_setup (struct clk *clk)
{
	printk (KERN_INFO "Panel: LMS606KF01\n");
	LMS606KF01_FUNC_TRACE
	
	bcm4760_set_pin_mux (vgpio_to_gpio(TT_VGPIO_LCD_ON),   BCM4760_PIN_MUX_GPIO);
	bcm4760_set_pin_mux (vgpio_to_gpio(TT_VGPIO_LCD_SCLK), BCM4760_PIN_MUX_GPIO);
	bcm4760_set_pin_mux (vgpio_to_gpio(TT_VGPIO_LCD_SDI),  BCM4760_PIN_MUX_GPIO);
	bcm4760_set_pin_mux (vgpio_to_gpio(TT_VGPIO_LCD_CS),   BCM4760_PIN_MUX_GPIO);
	bcm4760_set_pin_mux (vgpio_to_gpio(TT_VGPIO_LCD_SDO),  BCM4760_PIN_MUX_GPIO);

	gpio_request (TT_VGPIO_LCD_ON, "");
	gpio_request (TT_VGPIO_LCD_SCLK, "");
	gpio_request (TT_VGPIO_LCD_SDI, "");
	gpio_request (TT_VGPIO_LCD_CS, "");
	gpio_request (TT_VGPIO_LCD_SDO, "");

	gpio_direction_output (TT_VGPIO_LCD_ON, 0);
	gpio_direction_output (TT_VGPIO_LCD_SCLK, 1);
	gpio_direction_output (TT_VGPIO_LCD_SDI, 1);
	gpio_direction_output (TT_VGPIO_LCD_CS, 1);
	gpio_direction_input  (TT_VGPIO_LCD_SDO);
}

static struct clcd_panel_cntrl cntrl = {
	.enable		= lms606kf01_enable,
	.disable	= lms606kf01_disable,
	.suspend	= lms606kf01_suspend,
	.resume		= lms606kf01_resume,
	.init		= lms606kf01_init,
	.setup		= lms606kf01_setup
};

static struct clcd_panel timings = 
{
	.mode = {
		.name   	= LCM_MODEL_NAME,
		.refresh	= 60,   /* probably ignored, as it is bogus for all listed panels */
		.xres   	= PPL,
		.yres   	= LPP,
		.pixclock       = 90909,
	
		.left_margin    = HBP,  /* horizontal back porch */
		.right_margin   = HFP,  /* horizontal front porch */
		.upper_margin   = VBP,  /* vertical back porch */
		.lower_margin   = VFP,  /* vertical front porch */

		.hsync_len      = HSW,  /* horizontal pulse width */
		.vsync_len      = VSW,  /* vertical pulse width */
		.sync   	= 0,    /* horizontal and vertical both active low */
		.vmode  	= FB_VMODE_NONINTERLACED,
	},
	
	.width  = -1,
	.height = -1,
	.tim2   = PCD_HI << 27 | BCD << 26 | CPL << 16 | IOE << 14 | IPC << 13 | IHS << 12 | \
                  IVS << 11 | ACB << 6 | CLKSEL << 5 | PCD_LO << 0,
	.tim3   = 0,
	.cntl   = CNTL_LCDPWR | CNTL_LCDVCOMP(1) | CNTL_LCDTFT  | CNTL_LCDBPP16_565 | CNTL_BGR,
	.bpp    = 16,
};

static struct clcd_device lms606kf01 = {
	.timings= &timings,
	.cntrl	= &cntrl,
};

struct clcd_paneldev LMS606KF01_WithPLL =
{
	.panel 		= &lms606kf01,
	.lcd_mdiv	= 0x0000000A,
};

//arch_initcall(irvine_lcd_init);

#ifdef LMS606KF01_DEBUG

static inline void spi_read_data(unsigned char cmd, unsigned char * buf, int len)
{
	int i, j;
	unsigned char data;

	gpio_set_value(TT_VGPIO_LCD_CS, 0);
	
    /* control bit is 0 */
    gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
    gpio_set_value (TT_VGPIO_LCD_SDI, 0);
	gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
            
	for (i = 0; i < 8; i++) {
		gpio_set_value (TT_VGPIO_LCD_SCLK, 0);
		
		if (cmd & 0x80)
			gpio_set_value (TT_VGPIO_LCD_SDI, 1);
		else
			gpio_set_value (TT_VGPIO_LCD_SDI, 0);

		gpio_set_value (TT_VGPIO_LCD_SCLK, 1);
		cmd <<= 1;
	}

	/* Dummy clock cycle */
	gpio_set_value(TT_VGPIO_LCD_SCLK, 0);
	gpio_set_value(TT_VGPIO_LCD_SDI, 0);
	gpio_set_value(TT_VGPIO_LCD_SCLK, 1);

	/* read data*/
	for( i = 0; i < len; i++)
	{
		data = 0;
		for( j = 7; j >= 0; j--)
		{
			gpio_set_value(TT_VGPIO_LCD_SCLK, 0);
			gpio_set_value(TT_VGPIO_LCD_SCLK, 1);
			data |= (gpio_get_value(TT_VGPIO_LCD_SDO)<<j);
		}

		buf[i] = data;
	}

	gpio_set_value(TT_VGPIO_LCD_CS, 1);
}

static void level2_spi_read(unsigned char cmd, unsigned char *buf, int len)
{
	LMS606KF01_SET(password);
	LMS606KF01_SET(spi_read_en);
	spi_read_data(cmd, buf, len);
	LMS606KF01_SET(spi_read_dis);
}

static void level1_spi_read(unsigned char cmd, unsigned char *buf, int len)
{
	spi_read_data(cmd, buf, len);
}

static void lms606kf01_check_commands(void)
{	
	LMS606KF01_CHECK(smps_posi);
	LMS606KF01_CHECK(smps_nega);
	LMS606KF01_CHECK(smps);
	LMS606KF01_CHECK(power_control);
	LMS606KF01_CHECK(vcom);
	LMS606KF01_CHECK(source_control);
	LMS606KF01_CHECK(positive_r_gamma);
	LMS606KF01_CHECK(positive_g_gamma);
	LMS606KF01_CHECK(positive_b_gamma);
	LMS606KF01_CHECK(negative_r_gamma);
	LMS606KF01_CHECK(negative_g_gamma);
	LMS606KF01_CHECK(negative_b_gamma);
	LMS606KF01_CHECK(display_control);
	LMS606KF01_CHECK(resolution);
	LMS606KF01_CHECK(pixel_format);
}

static void lms606kf01_dump_level1(void)
{
	int i, j;
	unsigned char buf[8];
	unsigned char str_buf[128];
		
	for( i = 0; i < sizeof(level1_commands)/sizeof(struct lms606kf01_level1_command); i++)
	{
		if((level1_commands[i].rwc == R) && (level1_commands[i].para_size < sizeof(buf)))
		{
			level1_spi_read(level1_commands[i].opcode, buf, level1_commands[i].para_size);
			str_buf[0] = '\0';
			for( j = 0; j < level1_commands[i].para_size; j++)
			{ 
				unsigned char str_byte[8];
				snprintf(str_byte, sizeof(str_byte), "0x%x ", buf[j]);
				strcat(str_buf, str_byte);					
			}
			
			printk ("LMS606KF01 %s: %s %s\n", __FUNCTION__, level1_commands[i].str, str_buf);	
		}
	}	
}

#endif //LMS606KF01_DEBUG
